#include <linux/videodev.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-ioctl.h>
#include <linux/vmalloc.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/mm.h>

struct uservideo_mdev_info {
  int xsize, ysize;
  int framelen;
  int format;
} mdev_info;

static void *framebuf;
static int has_frame;
static wait_queue_head_t frameq;
static int activated;

static int uservideo_video_open( struct file *file ) {
  struct video_device *vdev = video_devdata(file);
  file->private_data = vdev;
  return 0;
}

static int uservideo_video_close( struct file *file ) {
  return 0;
}

static ssize_t uservideo_video_read(struct file *file, char *buf,
	size_t count, loff_t *ppos ) {
  return 0;
}

static unsigned int uservideo_video_poll(struct file *file, poll_table *wait) {
  return 0;
}

static long uservideo_video_do_ioctl( struct file *file,
	unsigned int cmd, void *arg );

static long uservideo_video_ioctl( struct file *file,
	unsigned int cmd, unsigned long arg ) {
  return video_usercopy(file, cmd, arg, uservideo_video_do_ioctl);
}

static long uservideo_video_do_ioctl( struct file *file,
	unsigned int cmd, void *arg ) {
  struct video_device *vdev = file->private_data;
  DECLARE_WAITQUEUE(wait, current);

  switch (cmd) {
  case VIDIOCGCAP: {
	struct video_capability *caps = arg;

	strcpy( caps->name, vdev->name );
	caps->type = VID_TYPE_CAPTURE;
	caps->channels = 1;
	caps->audios = 0;
	caps->minwidth = mdev_info.xsize;
	caps->minheight = mdev_info.ysize;
	caps->maxwidth = mdev_info.xsize;
	caps->maxheight = mdev_info.ysize;
	break;
  }
  case 	VIDIOCGMBUF: {
  	struct video_mbuf *vm = arg;

	memset( vm, 0, sizeof(struct video_mbuf) );
	vm->size = mdev_info.framelen;
	vm->frames = 1;
	vm->offsets[0] = 0;
	break;
  }
  case VIDIOCGWIN: {
	struct video_window *vw = arg;

	memset( vw, 0, sizeof(struct video_window) );
	vw->x = 0;
	vw->y = 0;
	vw->width = mdev_info.xsize;
	vw->height = mdev_info.ysize;
	vw->chromakey = 0;
	vw->flags = 0;
	break;
  }
  case VIDIOCSWIN: {
	struct video_window *vw = arg;

	if ( vw->width != mdev_info.xsize || vw->height != mdev_info.ysize )
		return -EINVAL;
	break;
  }
  case VIDIOCGPICT: {
	struct video_picture *vp = arg;

	memset( vp, 0, sizeof(struct video_picture) );
	vp->palette = mdev_info.format;
	break;
  }
  case VIDIOCSPICT: {
	struct video_picture *vp = arg;

	if ( vp->palette != mdev_info.format ) return -EINVAL;
	break;
  }
  case VIDIOCMCAPTURE:
  	break;
  case VIDIOCSYNC: {
  	int *frame = arg;

	if ( *frame != 0 ) return -EINVAL;
	add_wait_queue( &frameq, &wait );
	while ( !has_frame && !signal_pending(current) ) {
		set_current_state( TASK_INTERRUPTIBLE );
		schedule();
	}
	remove_wait_queue( &frameq, &wait );
	set_current_state( TASK_RUNNING );
	if ( signal_pending(current) ) return -ERESTARTSYS;

	has_frame = 0;
	break;
  }
  default:
  	return -ENOIOCTLCMD;
  }

  return 0;
}

static int uservideo_video_mmap( struct file *file, struct vm_area_struct *vma ) {
  int rc = remap_vmalloc_range( vma, framebuf, 0 );
  vma->vm_flags |= VM_DONTEXPAND;
  
  return rc;
}

static struct v4l2_file_operations uservideo_fops = {
  .owner =	THIS_MODULE,
  .open =	uservideo_video_open,
  .release =	uservideo_video_close,
  .read =	uservideo_video_read,
  .poll =	uservideo_video_poll,
  .mmap =	uservideo_video_mmap,
  .ioctl =	uservideo_video_ioctl,
};

static struct video_device uservideo_dev = {
  .name    = "User-space video driver",
  .fops    = &uservideo_fops,
  .release = video_device_release_empty,
};

int uservideo_proc_read2( struct file *f, char *buf, size_t count, loff_t *pos ) {
  char lbuf[200];
  int written, off = *pos;

  written = sprintf( lbuf, "active: %s\n", activated? "yes": "no" );
  written += sprintf( lbuf + written, "xsize: %d\n", mdev_info.xsize );
  written += sprintf( lbuf + written, "ysize: %d\n", mdev_info.ysize );
  written += sprintf( lbuf + written, "framelen: %d\n", mdev_info.framelen );
  written += sprintf( lbuf + written, "format: %d\n", mdev_info.format );
  if ( written < off ) {
  	written = 0;
	off -= written;
  }
  written -= off;
  if ( written > count ) {
  	if ( copy_to_user( buf + off, lbuf, count ) ) return -EFAULT;
	*pos += count;
	return count;
  }
  if ( written < 0 ) written = 0;

  if ( copy_to_user( buf + off, lbuf, written ) ) return -EFAULT;
  *pos += written;

  return written;
}

struct uservideo_dev_priv {
  char *buf;
  int bsize;
};

int uservideo_proc_mmap( struct file *f, struct vm_area_struct *vma ) {
  int rc = remap_vmalloc_range( vma, framebuf, 0 );
  vma->vm_flags |= VM_DONTEXPAND;
  
  return rc;
}

#define USERVIDEO_SET_MDEV_INFO	0x1301
#define USERVIDEO_ACTIVATE	0x1302
#define USERVIDEO_FRAME_READY	0x1303

int uservideo_proc_ioctl( struct inode *ind, struct file *f,
	unsigned int cmd, unsigned long arg ) {
  struct uservideo_mdev_info old_info;
  int activate;

  switch ( cmd ) {
  case USERVIDEO_SET_MDEV_INFO:
	old_info = mdev_info;
	if ( copy_from_user( &mdev_info, (void*) arg, sizeof(mdev_info) ) )
		return -EFAULT;
	if ( old_info.framelen != mdev_info.framelen ) {
	  	vfree( framebuf );
		framebuf = vmalloc_user( mdev_info.framelen );
	}
	break;
  case USERVIDEO_ACTIVATE:
	if ( copy_from_user( &activate, (void*) arg, sizeof(activate) ) )
		return -EFAULT;
	if ( activate && !activated ) {
		int i, video_nr = -1;

  		i = video_register_device( &uservideo_dev, VFL_TYPE_GRABBER, video_nr );
		if ( i < 0 ) {
			printk(KERN_ERR "Failed to register as video device (%d).\n", i );
			return -EFAULT;
		}
  
		printk(KERN_INFO __FILE__": registered as /dev/video%d.\n",
  			uservideo_dev.minor & 0x3F);
	}
	else if ( !activate && activated )
		video_unregister_device( &uservideo_dev );
	activated = activate;
  	break;
  case USERVIDEO_FRAME_READY:
  	has_frame = 1;
  	wake_up_interruptible( &frameq );
	break;
  default:
  	return -ENOIOCTLCMD;
  }

  return 0;
}

struct proc_dir_entry *uservideo_proc;

struct file_operations uservideo_pfops = {
read: uservideo_proc_read2,
mmap: uservideo_proc_mmap,
ioctl: uservideo_proc_ioctl,
};

int __init uservideo_init(void) {
  mdev_info.xsize = 320;
  mdev_info.ysize = 240;
  mdev_info.framelen = 320 * 240 * 3;
  mdev_info.format = VIDEO_PALETTE_RGB24;

  framebuf = vmalloc_user( mdev_info.framelen );

  uservideo_proc = create_proc_entry( "driver/user_video", 0444, 0 );
  uservideo_proc->proc_fops = &uservideo_pfops;

  has_frame = 0;
  init_waitqueue_head( &frameq );
  activated = 0;

  return 0;
}

void __exit uservideo_exit(void) {
  if ( activated ) video_unregister_device( &uservideo_dev );
  remove_proc_entry( "driver/user_video", 0 );
  vfree( framebuf );
}

MODULE_DESCRIPTION("User-space video driver");
MODULE_AUTHOR("Vad Rulezz, <vr5@narod.ru>");
MODULE_LICENSE("GPL");

module_init( uservideo_init );
module_exit( uservideo_exit );
